/*    This file is part of the Gudhi Library - https://gudhi.inria.fr/ - which
 * is released under MIT. See file LICENSE or go to
 * https://gudhi.inria.fr/licensing/ for full license details. Author(s): David

 *
 *    Copyright (C) 2023 Inria
 *
 *    Modification(s):
 *      - YYYY/MM Author: Description of the modification
 */

#ifndef FINITELY_CRITICAL_FILTRATIONS_H_
#define FINITELY_CRITICAL_FILTRATIONS_H_

#include <algorithm>
#include <cmath>
#include <cstddef>
#include <cstdint>
#include <functional>
#include <iostream>
#include <limits>
#include <vector>

namespace Gudhi::multiparameter::multi_filtrations {

/** \class Finitely_critical_multi_filtration
 * Finitely_critical_multi_filtration.h
 * gudhi/Simplex_tree/multi_filtrations/Finitely_critical_multi_filtration.h
 * \brief Vector-like filtration value, for multiparameter persistence, with
 * numpy-like methods.
 *
 * \ingroup multiparameter
 *
 * \details Child of `std::vector<T>` that has numpy-like pointwise operators.
 * One critical simplicial filtrations are filtrations such that the lifetime
 * each simplex is a positive cone, e.g.
 *  - \f$ \{ x \in  \mathbb R^2 : x>=(1,2)\} \f$ is valid, while
 *  - \f$ \{ x \in  \mathbb R^2 : x>=(1,2)\} \cap \{x \in  \mathbb R^2 :
 * x>=(2,1)\} \f$ is not- Finitely critical filtrations are filtrations such
 * that the lifetime of each simplex is a union of such cones, e.g.,
 *  - \f$ \{ x \in  \mathbb R^2 : x>=(1,2)\} \cap \{ x \in  \mathbb R^2 :
 * x>=(2,1)\} \f$ is finitely critical, and more particularly 2-critical, while
 *  - \f$ \{ x \in  \mathbb R^2 : x>= \mathrm{epigraph}(y\mapsto e^{-y})\} \f$
 * is not. Non 1-critical filtrations are not supported yet. \tparam T value
 * type of the vector-like.
 */
template <typename T>
class Finitely_critical_multi_filtration : public std::vector<T> {
public:
  Finitely_critical_multi_filtration() : std::vector<T>(){};
  Finitely_critical_multi_filtration(int n)
      : std::vector<T>(n, -T_inf){}; // minus infinity by default
  Finitely_critical_multi_filtration(int n, T value)
      : std::vector<T>(n, value){};
  Finitely_critical_multi_filtration(std::initializer_list<T> init)
      : std::vector<T>(init){};
  Finitely_critical_multi_filtration(const std::vector<T> &v)
      : std::vector<T>(v){};
  Finitely_critical_multi_filtration(std::vector<T> &&v)
      : std::vector<T>(std::move(v)){};
  Finitely_critical_multi_filtration(typename std::vector<T>::iterator it_begin,
                                     typename std::vector<T>::iterator it_end)
      : std::vector<T>(it_begin, it_end){};
  Finitely_critical_multi_filtration(
      typename std::vector<T>::const_iterator it_begin,
      typename std::vector<T>::const_iterator it_end)
      : std::vector<T>(it_begin, it_end){};

  using std::vector<T>::operator[];
  using value_type = T;
  using OneCritical = Finitely_critical_multi_filtration<T>;
  template <typename value_type>
  using Base = Finitely_critical_multi_filtration<value_type>;
  operator std::vector<T> &() const { return *this; }
  std::vector<T> get_vector() const {
    return static_cast<std::vector<T>>(*this);
  }
  std::size_t num_parameters() const { return this->size(); }
  // TODO : REMOVE SIZE ( confusing when kcritical) 
  // static std::size_t num_generators() {return 1;}

  operator std::vector<T>() const { return static_cast<std::vector<T>>(*this); }

  inline const std::vector<T> & as_vector() const { return *this; }

  inline bool is_inf() const {
    if (this->size() != 1)
      return false;
    return (this->operator[](0) == T_inf);
  }
  inline bool is_minus_inf() const {
    if constexpr (std::is_same<T, bool>::value) {
      return false; // suppresses a warning
    } else {
      if (this->size() != 1)
        return false;
      return (this->operator[](0) == -T_inf);
    }
  }
  inline bool is_nan() const {
    if (this->size() != 1)
      return false;
    return std::isnan(*(this->begin()));
  }
  inline bool is_finite() const {
    if (this->size() > 1)
      return true;
    if (this->size() == 0)
      return false;
    auto first_value = *(this->begin()); // TODO : Maybe check all entries ?
    if (std::isnan(first_value) || first_value == -T_inf ||
        first_value == T_inf)
      return false;
    return true;
  }

  inline friend bool operator<(const Finitely_critical_multi_filtration<T> &a,
                               const Finitely_critical_multi_filtration<T> &b) {
    if (a.is_inf() || a.is_nan() || b.is_nan() || b.is_minus_inf())
      return false;
    if (b.is_inf() || a.is_minus_inf())
      return true;
    bool isSame = true;
    auto n = a.size();
    assert(a.size() == b.size());
    for (auto i = 0u; i < n; ++i) {
      if (a[i] > b[i])
        return false;
      if (isSame && a[i] != b[i])
        isSame = false;
    }
    if (isSame)
      return false;
    return true;
  }
  inline friend bool operator<=(const Finitely_critical_multi_filtration<T> &a,
                                const Finitely_critical_multi_filtration<T> &b) {
    if (a.is_nan() || b.is_nan())
      return false;
    if (b.is_inf() || a.is_minus_inf())
      return true;
    if (a.is_inf() || b.is_minus_inf())
      return false;
    auto n = a.size();
    assert(a.size() == b.size());
    for (std::size_t i = 0u; i < n; ++i) {
      if (a[i] > b[i])
        return false;
    }
    return true;
  }

  // GREATER THAN OPERATORS
  inline friend bool operator>(const Finitely_critical_multi_filtration<T> &a,
                               const Finitely_critical_multi_filtration<T> &b) {
    return b < a;
  }
  inline friend bool operator>=(const Finitely_critical_multi_filtration<T> &a,
                                const Finitely_critical_multi_filtration<T> &b) {
    return b <= a;
  }

  inline Finitely_critical_multi_filtration<T> &
  operator=(const Finitely_critical_multi_filtration<T> &a) {
    std::vector<T>::operator=(a);
    return *this;
  }
  inline Finitely_critical_multi_filtration<T>& operator-() {
    for (auto& truc : *this) {
      truc = -truc;
    }
  }
  inline Finitely_critical_multi_filtration<T> copy() const { return *this; }

  std::vector<T> &_convert_back() { return *this; }
  constexpr static Finitely_critical_multi_filtration<T> inf() {
    return {T_inf};
  }
  constexpr static Finitely_critical_multi_filtration<T> minus_inf() {
    return {-T_inf};
  }
  constexpr static Finitely_critical_multi_filtration<T> nan() {
    return {std::numeric_limits<T>::quiet_NaN()};
  }

  // operators *=
  inline friend Finitely_critical_multi_filtration<T> &
  operator-=(Finitely_critical_multi_filtration<T> &result,
             const Finitely_critical_multi_filtration<T> &to_substract) {
    if (result.is_nan() || to_substract.is_nan() ||
        (result.is_inf() && to_substract.is_inf()) ||
        (result.is_minus_inf() && to_substract.is_minus_inf())) [[unlikely]] {
      result = std::numeric_limits<T>::quiet_NaN();
      return result;
    }
    if (result.is_inf() || to_substract.is_minus_inf()) [[unlikely]] {
      result = inf();
      return result;
    }
    if (result.is_minus_inf() || to_substract.is_inf()) [[unlikely]] {
      result = minus_inf();
      return result;
    }
    std::transform(result.begin(), result.end(), to_substract.begin(),
                   result.begin(), std::minus<T>());
    return result;
  }
  inline friend Finitely_critical_multi_filtration<T> &
  operator+=(Finitely_critical_multi_filtration<T> &result,
             const Finitely_critical_multi_filtration<T> &to_add) {

    if (result.is_nan() || to_add.is_nan() ||
        (result.is_inf() && to_add.is_minus_inf()) ||
        (result.is_minus_inf() && to_add.is_inf())) [[unlikely]] {
      result = nan();
      return result;
    }
    if (result.is_inf() || to_add.is_inf()) [[unlikely]] {
      result = inf();
      return result;
    }
    if (result.is_minus_inf() || to_add.is_minus_inf()) [[unlikely]] {
      result = minus_inf();
      return result;
    }

    std::transform(result.begin(), result.end(), to_add.begin(), result.begin(),
                   std::plus<T>());
    return result;
  }
  inline friend Finitely_critical_multi_filtration<T> &
  operator*=(Finitely_critical_multi_filtration<T> &result,
             const Finitely_critical_multi_filtration<T> &to_add) {

    assert(result.is_finite() || to_add.is_finite());
    std::transform(result.begin(), result.end(), to_add.begin(), result.begin(),
                   std::multiplies<T>());
    return result;
  }

  inline friend Finitely_critical_multi_filtration<T> &
  operator-=(Finitely_critical_multi_filtration<T> &result,
             const T &to_substract) {
    for (auto &truc : result) {
      truc -= to_substract;
    }
    return result;
  }
  inline friend Finitely_critical_multi_filtration<T> &
  operator+=(Finitely_critical_multi_filtration<T> &result, const T &to_add) {
    for (auto &truc : result) {
      truc += to_add;
    }
    return result;
  }
  inline friend Finitely_critical_multi_filtration<T> &
  operator*=(Finitely_critical_multi_filtration<T> &result, const T &to_add) {
    if (to_add == T_inf || to_add == -T_inf) [[unlikely]] {
      for (auto &truc : result) {
        if (truc > 0)
          truc = to_add;
        else if (truc < 0)
          truc = -to_add;
        // 0*inf = 0
      }
      return result;
    }
    for (auto &truc : result) {
      truc *= to_add;
    }
    return result;
  }
  inline friend Finitely_critical_multi_filtration<T> &
  operator/=(Finitely_critical_multi_filtration<T> &result, const T &to_add) {
    for (auto &truc : result) {
      truc /= to_add;
    }
    return result;
  }

  /// OPERATORS *
  inline friend Finitely_critical_multi_filtration<T>
  operator+(Finitely_critical_multi_filtration<T> result, const T &to_add) {
    result += to_add;
    return result;
  }

  inline friend Finitely_critical_multi_filtration<T>
  operator+(Finitely_critical_multi_filtration<T> result,
            const Finitely_critical_multi_filtration<T> &to_add) {
    result += to_add;
    return result;
  }
  inline friend Finitely_critical_multi_filtration<T>
  operator-(Finitely_critical_multi_filtration<T> result,
            const Finitely_critical_multi_filtration<T> &to_add) {
    result -= to_add;
    return result;
  }
  inline friend Finitely_critical_multi_filtration<T>
  operator-(Finitely_critical_multi_filtration<T> result, const T &to_add) {
    result -= to_add;
    return result;
  }
  inline friend Finitely_critical_multi_filtration<T>
  operator*(Finitely_critical_multi_filtration<T> result, const T &to_add) {
    result *= to_add;
    return result;
  }

  inline friend Finitely_critical_multi_filtration<T>
  operator+(const T &to_add,
            const Finitely_critical_multi_filtration<T> &result) {
    return result + to_add;
  }
  inline friend Finitely_critical_multi_filtration<T>
  operator-(const T &to_add,
            const Finitely_critical_multi_filtration<T> &result) {
    return result - to_add;
  }
  inline friend Finitely_critical_multi_filtration<T>
  operator*(const T &to_add,
            const Finitely_critical_multi_filtration<T> &result) {
    return result * to_add;
  }

  // template<class array_like>
  inline friend bool
  operator==(const Finitely_critical_multi_filtration<T> &self,
             const Finitely_critical_multi_filtration<T> &to_compare) {
    if (self.num_parameters() != to_compare.num_parameters())
      return false;
    auto it = to_compare.begin();
    for (auto i = 0u; i < self.num_parameters(); i++) {
      if (self.at(i) != *(it++))
        return false;
    }
    return true;
  }

  inline static std::vector<std::vector<T>> to_python(
      const std::vector<Finitely_critical_multi_filtration<T>> &to_convert) {
    return std::vector<std::vector<T>>(to_convert.begin(), to_convert.end());
  }

  inline static std::vector<Finitely_critical_multi_filtration<T>>
  from_python(const std::vector<std::vector<T>> &to_convert) {
    return std::vector<Finitely_critical_multi_filtration<T>>(
        to_convert.begin(), to_convert.end());
  }

  /** \brief This functions take the filtration value `this` and pushes it to
   * the cone \f$ \{ y\in \mathbb R^n : y>=x \} \f$. After calling this method,
   * the value of this is updated to \f$ \mathrm{this} = \min \{ y\in \mathbb
   * R^n : y>=this \}\cap \{ y\in \mathbb R^n : y>=x \}
   * @param[in] x The target filtration value on which to push `this`.
   */
  inline void push_to(const Finitely_critical_multi_filtration<T> &x) {
    if (this->is_inf() || this->is_nan() || x.is_nan() || x.is_minus_inf())
      return;
    if (x.is_inf() || this->is_minus_inf()) {
      *this = x;
      return;
    }
    if (this->num_parameters() != x.num_parameters()) {
      std::cerr << "Does only work with 1-critical filtrations ! Sizes "
                << this->num_parameters() << " and " << x.num_parameters() << "are different !"
                << std::endl;
      std::cerr << "This : " << *this << std::endl;
      std::cerr << "arg : " << x << std::endl;
      throw std::logic_error("Bad sizes");
    }
    for (std::size_t i = 0; i < x.num_parameters(); i++)
      this->operator[](i) =
          this->operator[](i) > x[i] ? this->operator[](i) : x[i];
  }

  /** \brief This functions take the filtration value `this` and pulls it to the
   * cone \f$ \{ y\in \mathbb R^n : y<=x \} \f$. After calling this method, the
   * value of this is updated to \f$ \mathrm{this} = \max \{ y\in \mathbb R^n :
   * y<=this \}\cap \{ y\in \mathbb R^n : y<=x \}
   * @param[in] x The target filtration value on which to push `this`.
   */
  inline void pull_to(const Finitely_critical_multi_filtration<T> &x) {
    if (x.is_inf() || this->is_nan() || x.is_nan() || this->is_minus_inf())
      return;
    if (this->is_inf() || x.is_minus_inf()) {
      *this = x;
      return;
    }
    if (this->num_parameters() != x.num_parameters()) {
      std::cerr << "Does only work with 1-critical filtrations ! Sizes "
                << this->num_parameters() << " and " << x.num_parameters() << "are different !"
                << std::endl;
      std::cerr << "This : " << *this << std::endl;
      std::cerr << "arg : " << x << std::endl;
      throw std::logic_error("Bad sizes");
    }
    for (std::size_t i = 0u; i < x.num_parameters(); i++)
      this->operator[](i) =
          this->operator[](i) > x[i] ? x[i] : this->operator[](i);
  }
  // Warning, this function  assumes that the comparisons checks have already
  // been made !
  inline void insert_new(Finitely_critical_multi_filtration<T> to_concatenate) {
    this->insert(this->end(), std::move_iterator(to_concatenate.begin()),
                 std::move_iterator(to_concatenate.end()));
  }

  // scalar product of a filtration value with x.
  template <typename U=T>
  inline U linear_projection(const std::vector<U> &x) {
    U projection = 0;
    std::size_t size = std::min(x.size(), this->size());
    for (std::size_t i = 0u; i < size; i++)
      projection += x[i] * static_cast<U>(this->operator[](i));
    return projection;
  }

  // easy debug
  inline friend std::ostream &
  operator<<(std::ostream &stream,
             const Finitely_critical_multi_filtration<T> &truc) {
    if (truc.is_inf()) {
      stream << "[inf, ..., inf]";
      return stream;
    }
    if (truc.is_minus_inf()) {
      stream << "[-inf, ..., -inf]";
      return stream;
    }
    if (truc.is_nan()) {
      stream << "[NaN]";
      return stream;
    }
    if (truc.empty()) {
      stream << "[]";
      return stream;
    }
    stream << "[";
    for (std::size_t i = 0; i < truc.size() - 1; i++) {
      stream << truc[i] << ", ";
    }
    if (!truc.empty())
      stream << truc.back();
    stream << "]";
    return stream;
  }
  inline T norm() const {
    T out = 0;
    for (auto &stuff : *this)
      out += std::pow(stuff, 2);
    return sqrt(out);
  }
  inline T distance(const Finitely_critical_multi_filtration<T> &other) const {
    T out = 0;
    for (std::size_t i = 0u; i < other.size(); i++) {
      out += std::pow((*this)[i] - other[i], 2);
    }
    return sqrt(out);
  }
  template <typename out_type=std::int32_t,typename U>
  inline Finitely_critical_multi_filtration<out_type>
  coordinates_in_grid(const std::vector<std::vector<U>>& grid) const {
    Finitely_critical_multi_filtration<out_type> coords(this->size());
    assert(grid.size() >= this->size());
    // U grid_inf = std::numeric_limits<U>::has_infinity
    //                  ? std::numeric_limits<U>::infinity()
    //                  : std::numeric_limits<U>::max();
    for (std::size_t parameter = 0u; parameter < grid.size(); ++parameter) {
      const auto &filtration = grid[parameter];
      auto C = std::distance(
          filtration.begin(),
          std::lower_bound(filtration.begin(), filtration.end(),
                           static_cast<U>(this->operator[](parameter))));
      coords[parameter] = static_cast<out_type>(C);
    }
    return coords;
  }
  template <typename U>
  inline Finitely_critical_multi_filtration<U>
  evaluate_in_grid(const std::vector<std::vector<U>>& grid) const {
    Finitely_critical_multi_filtration<U> pushed_value(this->size());
    assert(grid.size() == this->size());
    U grid_inf = std::numeric_limits<U>::has_infinity
                     ? std::numeric_limits<U>::infinity()
                     : std::numeric_limits<U>::max();
    for (std::size_t parameter = 0u; parameter < grid.size(); ++parameter) {
      const auto &filtration = grid[parameter];
      const auto& c = this->operator[](parameter);
      pushed_value[parameter] = c == T_inf ? grid_inf : filtration[c];
    }
    return pushed_value;
  }

  template <typename oned_array>
  inline void coordinates_in_grid_inplace(const std::vector<oned_array>& grid,
                                          bool coordinate = true) {
    Finitely_critical_multi_filtration<std::size_t> coords(this->size());
    assert(grid.size() >= this->size());
    for (std::size_t parameter = 0u; parameter < grid.size(); ++parameter) {
      const auto &filtration = grid[parameter];
      auto d =
          std::distance(filtration.begin(),
                        std::lower_bound(filtration.begin(), filtration.end(),
                                         this->operator[](parameter)));
      this->operator[](parameter) =
          coordinate ? static_cast<T>(d) : static_cast<T>(grid[parameter][d]);
      ;
    }
  }
  template <typename U> inline Finitely_critical_multi_filtration<U> astype() const {
    Finitely_critical_multi_filtration<U> out(this->size());
    for (std::size_t i = 0u; i < this->size(); i++)
      out[i] = static_cast<U>(this->operator[](i));
    return out;
  }

public:
  // TODO : maybe add the {inf}, minus inf, nan  there as static members? this
  // would make comparisons faster (just compare the ptr)
  // TODO : I'm not sure why constexpr doens't work anymore
  constexpr static const T T_inf = std::numeric_limits<T>::has_infinity
                                       ? std::numeric_limits<T>::infinity()
                                       : std::numeric_limits<T>::max();

  // for compiler
  constexpr static bool is_multi_critical = false;
};

template <typename T, bool co=false> class KCriticalFiltration {
public:
  using value_type = T;
  using OneCritical = Finitely_critical_multi_filtration<T>;
  template <typename value_type>
  using Base = KCriticalFiltration<value_type>;
  std::size_t num_parameters() const { return multi_filtration_.size() == 0 ? 0 : multi_filtration_[0].size(); }
  KCriticalFiltration() : multi_filtration_(1)
  {
  multi_filtration_[0] = co ? Finitely_critical_multi_filtration<T>::inf() : Finitely_critical_multi_filtration<T>::minus_inf();
  }; // initialize at 1, with 0 parameter (consistent with below)
  KCriticalFiltration(int n)
      : multi_filtration_({Finitely_critical_multi_filtration<T>(
            n)}){}; // minus infinity by default
  KCriticalFiltration(int n, T value)
      : multi_filtration_({Finitely_critical_multi_filtration<T>(
            n, value)}){}; // std::vector<T>(n, value){};
  KCriticalFiltration(std::initializer_list<T> init)
      : multi_filtration_({Finitely_critical_multi_filtration<T>(
            init)}){}; // : std::vector<T>({std::vector<T>(init)};
  KCriticalFiltration(const std::vector<T> &v) : multi_filtration_({v}){};
  KCriticalFiltration(std::vector<T> &&v) : multi_filtration_({std::move(v)}){};
  KCriticalFiltration(const std::vector<Finitely_critical_multi_filtration<T>> &v) : multi_filtration_(v){};
  KCriticalFiltration(std::vector<Finitely_critical_multi_filtration<T>> &&v) : multi_filtration_(std::move(v)){};
  KCriticalFiltration(typename std::vector<T>::iterator it_begin,
                      typename std::vector<T>::iterator it_end)
      : multi_filtration_(
            {Finitely_critical_multi_filtration<T>(it_begin, it_end)}){};
  KCriticalFiltration(typename std::vector<T>::const_iterator it_begin,
                      typename std::vector<T>::const_iterator it_end)
      : multi_filtration_({Finitely_critical_multi_filtration<T>(
            it_begin, it_end)}){}; // : std::vector<T>(it_begin, it_end){};
  KCriticalFiltration(const KCriticalFiltration<T> &other)
      : multi_filtration_(other.multi_filtration_){};

  KCriticalFiltration &operator=(const KCriticalFiltration<T> &other) {
    this->multi_filtration_ = other.multi_filtration_;
    return *this;
  }
  inline friend bool
  operator==(const KCriticalFiltration<T,co> &self,
             const KCriticalFiltration<T,co> &to_compare) {
    if (self.num_generators() != to_compare.num_generators())
      return false;
    auto it = to_compare.begin();
    for (auto i = 0u; i < self.num_generators(); i++) {
      if (self[i] != *(it++))
        return false;
    }
    return true;
  }
  void reserve(std::size_t n) { this->multi_filtration_.reserve(n); }
  void set_num_generators(std::size_t n) {this->multi_filtration_.resize(n);}

  inline bool is_inf() const {
    return this->multi_filtration_.size() == 1 &&
           this->multi_filtration_[0].is_inf();
  }
  inline bool is_minus_inf() const {
    return this->multi_filtration_.size() == 1 &&
           this->multi_filtration_[0].is_minus_inf();
  }
  inline bool is_nan() const {
    return this->multi_filtration_.size() == 1 &&
           this->multi_filtration_[0].is_nan();
  }
  inline bool is_finite() const {
    if (this->empty())
      return false;
    for (auto &stuff : *this) {
      if (!stuff.is_finite())
        return false;
    }
    return true;
  }

  operator OneCritical() const {
    assert(this->num_generators() == 1);
    return this->multi_filtration_[0];
  }

  OneCritical &operator[](std::size_t i) { return this->multi_filtration_[i]; }
  inline OneCritical factorize_below() const {
    if (this->num_generators() == 0) [[unlikely]]
      return OneCritical();
    OneCritical result(multi_filtration_[0].num_parameters(), OneCritical::T_inf);
    for (const auto &stuff : this->multi_filtration_) {
      if (stuff.is_nan() || stuff.is_minus_inf())
        return stuff;
      if (stuff.is_inf())
        continue;
      for (std::size_t i = 0; i < stuff.num_parameters(); ++i) {
        result[i] = std::min(result[i], stuff[i]);
      }
    }
    return result;
  }

  inline OneCritical factorize_above() const {
    if (this->num_generators() == 0) [[unlikely]]
      return OneCritical();
    OneCritical result(multi_filtration_[0].num_parameters(), -OneCritical::T_inf);
    for (auto &stuff : this->multi_filtration_) {
      if (stuff.is_nan() || stuff.is_inf())
        return stuff;
      if (stuff.is_minus_inf())
        continue;
      for (std::size_t i = 0; i < stuff.num_parameters(); ++i) {
        result[i] = std::max(result[i], stuff[i]);
      }
    }
    return result;
  }
  /** \brief This functions take the filtration value `this` and pushes it to
   * the cone \f$ \{ y\in \mathbb R^n : y>=x \} \f$. After calling this method,
   * the value of this is updated to \f$ \mathrm{this} = \min \{ y\in \mathbb
   * R^n : y>=this \}\cap \{ y\in \mathbb R^n : y>=x \}
   * @param[in] x The target filtration value on which to push `this`.
   */
  inline void push_to(const Finitely_critical_multi_filtration<T> &x) {
    if (this->is_inf() || this->is_nan() || x.is_nan() || x.is_minus_inf())
      return;
    if (x.is_inf() || this->is_minus_inf()) {
      *this = x;
      return;
    }
    for (auto &stuff : *this) {
      stuff.push_to(x);
    }
  }
  // TODO : this is not well defined for kcritical <-> kcritical

  /** \brief This functions take the filtration value `this` and pulls it to the
   * cone \f$ \{ y\in \mathbb R^n : y<=x \} \f$. After calling this method, the
   * value of this is updated to \f$ \mathrm{this} = \max \{ y\in \mathbb R^n :
   * y<=this \}\cap \{ y\in \mathbb R^n : y<=x \}
   * @param[in] x The target filtration value on which to push `this`.
   */
  inline void pull_to(const Finitely_critical_multi_filtration<T> &x) {
    if (x.is_inf() || this->is_nan() || x.is_nan() || this->is_minus_inf())
      return;
    if (this->is_inf() || x.is_minus_inf()) {
      *this = x;
      return;
    }
    for (auto &stuff : *this) {
      stuff.pull_to(x);
    }
  } 
  // cannot be const, as gudhi iterators are not const
  template <typename U> inline U linear_projection(const std::vector<U> &x) {
    if constexpr (co) {
      U projection = std::numeric_limits<U>::lowest();
      for (const auto &y : *this) {
        projection = std::max(projection, y.linear_projection(x));
      }
      return projection;
    } else {
      U projection = std::numeric_limits<U>::max();
      for (auto &y : *this) { // cannot be const (Gudhi)
        projection = std::min(projection, y.linear_projection(x));
      }
      return projection;
    }
  }

  /*
  * Checks if b is cleanable with respect to a
  */
  static inline bool dominates(const OneCritical&a, const OneCritical&b, value_type max_error) {
    if constexpr (co)
      return a - max_error <= b;
    else {
      return a + max_error >= b;
    }
  }

  static inline bool dominates(const OneCritical&a, const OneCritical&b) {
    if constexpr (co)
      return a <= b;
    else {
      return a >= b;
    }
  }
  static inline bool strictly_dominates(const OneCritical&a, const OneCritical&b) {
    if constexpr (co)
      return a < b;
    else {
      return a > b;
    }
  }

  template <typename U>
  inline KCriticalFiltration<U>
  evaluate_in_grid(const std::vector<std::vector<U>>& grid) const {
    KCriticalFiltration<U> out(this->num_generators());
    for (std::size_t i = 0; i < this->num_generators(); ++i){
      out[i] = this->operator[](i).evaluate_in_grid(grid);
    }
    return out;
  }
  
  inline void clean(value_type max_error = 0) {
    // A bit ugly, todo : erase+removeif ?
    for (std::size_t i = 0; i < multi_filtration_.size(); i++) {
      for (std::size_t j = 0; j < multi_filtration_.size(); j++) {
        if (i == j)
          continue;
        if (dominates(multi_filtration_[j], multi_filtration_[i], max_error)) {
          multi_filtration_[i].clear();
        }
        while (j < multi_filtration_.size() && dominates(multi_filtration_[i], multi_filtration_[j], max_error)) {
          multi_filtration_[j].clear();
          i--;
        }
      }
    }
    multi_filtration_.erase(
        std::remove_if(multi_filtration_.begin(), multi_filtration_.end(),
                       [](const Finitely_critical_multi_filtration<T> &a) {
                         return a.empty();
                       }),
        multi_filtration_.end());
  }
  inline void add_point(const Finitely_critical_multi_filtration<T> &x) {
    const bool verbose = false;
    if (multi_filtration_.empty()) {
      if constexpr (verbose)
        std::cout << "Adding x=" << x << " (currently empty)" << std::endl;
      multi_filtration_.push_back(x);
      return;
    }
    for (const auto &y : multi_filtration_){
      if (dominates(x,y)){
        return;
      }
    }
    if constexpr (verbose)
      std::cout << "x: " << x << " is useful, removing unnecessary entries"
                << std::endl;
    multi_filtration_.erase(
        std::remove_if(multi_filtration_.begin(), multi_filtration_.end(),
                       [&x](const Finitely_critical_multi_filtration<T> &y) {
                         if constexpr (verbose) {
                           if (dominates(y, x)) {
                             std::cout << "Removing y=" << y << std::endl;
                           } else {
                             std::cout << "Keeping y=" << y << std::endl;
                           }
                         }
                         return dominates(y, x);
                       }),
        multi_filtration_.end());

    multi_filtration_.push_back(x);
  }
  inline void re_clean(){
    // Ensures all points are useful again. Can be useful if points are added manually.
    // TODO : maybe optimize
    KCriticalFiltration<value_type> out; // should be inf
    out.multi_filtration_.reserve(this->multi_filtration_.size());
    for (const auto& x : multi_filtration_){
      out.add_point(x);
    }
    std::swap(multi_filtration_, out);
  }

  // easy debug
  inline friend std::ostream &operator<<(std::ostream &stream,
                                         const KCriticalFiltration<T,co> &truc) {
    if (truc.is_inf()) {
      stream << "[inf, ..., inf]";
      return stream;
    }
    if (truc.is_minus_inf()) {
      stream << "[-inf, ..., -inf]";
      return stream;
    }
    if (truc.is_nan()) {
      stream << "[NaN]";
      return stream;
    }
    stream << "(k=" << truc.multi_filtration_.size() << ")[";
    for (const auto &machin : truc) {
      stream << machin << "; ";
    }
    if (truc.multi_filtration_.size() > 0) {
      stream << "\b"
             << "\b";
    }
    stream << "]";
    return stream;
  }
  inline void clear() { multi_filtration_.clear(); }
  inline bool empty() const { return multi_filtration_.empty(); }
  inline const OneCritical& operator[] (std::size_t i) const { return multi_filtration_[i]; }

  inline typename std::vector<Finitely_critical_multi_filtration<T>>::iterator
  begin() {
    return multi_filtration_.begin();
  }
  inline typename std::vector<Finitely_critical_multi_filtration<T>>::iterator
  end() {
    return multi_filtration_.end();
  }
  inline typename std::vector<
      Finitely_critical_multi_filtration<T>>::const_iterator
  begin() const {
    return multi_filtration_.begin();
  }
  inline typename std::vector<
      Finitely_critical_multi_filtration<T>>::const_iterator
  end() const {
    return multi_filtration_.end();
  }
  template <typename oned_array>
  inline KCriticalFiltration<std::int32_t>
  coordinates_in_grid(const std::vector<oned_array>& grid) const {
    assert(grid.size() >= this->num_generators());
    KCriticalFiltration<std::int32_t> out(this->num_generators());
    for (std::size_t i = 0u; i < this->num_generators(); ++i) {
      out[i] = this->multi_filtration_[i].coordinates_in_grid(grid);
    }
    return out;
  }
  template <typename oned_array>
  inline void coordinates_in_grid_inplace(const std::vector<oned_array>& grid,
                                          bool coordinate = true) {
    assert(grid.size() >= this->num_generators());
    for (auto &x : this->multi_filtration_) {
      x.coordinates_in_grid_inplace(grid, coordinate);
    }
  }
  template <typename U> inline KCriticalFiltration<U> astype() const {
    std::vector<Finitely_critical_multi_filtration<U>> out(this->num_generators());
    for (std::size_t i = 0u; i < this->num_generators(); ++i) {
      out[i] = this->multi_filtration_[i].template astype<U>();
    }
    return KCriticalFiltration<U>(std::move(out));
  }
  inline void push_back(const OneCritical& x){
    multi_filtration_.push_back(x);
  }
  inline const std::vector<Finitely_critical_multi_filtration<T>>& as_vector() const {
    return multi_filtration_;
  }

  inline std::vector<std::vector<T>> as_VECTOR() const {
    std::vector<std::vector<T>> out(this->num_generators(),
                                    std::vector<T>(this->num_parameters()));
    for (std::size_t i = 0; i < this->num_generators(); ++i) {
      for (std::size_t j = 0; j < this->num_parameters(); ++j) {
        out[i][j] = multi_filtration_[i][j];
      }
    }
    return out;
  }

  inline void _clean(bool keep_inf=true) {
    multi_filtration_.erase(std::remove_if(multi_filtration_.begin(), multi_filtration_.end(),
                              [keep_inf](const OneCritical &a) {
                                return a.empty() ||
                                       ((!keep_inf) &&
                                        (a.is_inf() || a.is_minus_inf()));
                              }),
                            multi_filtration_.end());
  }
  inline std::size_t num_generators() const {
    return multi_filtration_.size();
  }



  // TODO : this costs a lot... optimize / cheat in some way for python ?
  inline bool operator<(const KCriticalFiltration<T,co> &other) const {
    //check if this curves is below other's curve 
    // ie for each guy in this, check if there is a guy in other that dominates him
    for (std::size_t i = 0u; i < multi_filtration_.size(); ++i) {
      for (std::size_t j = 0u; j < other.multi_filtration_.size(); ++j) {
        // i<j
        if (strictly_dominates(other.multi_filtration_[j],multi_filtration_[i]))
          continue;
      }
      return false;
    }
    return true;
  }
  inline bool operator>(const KCriticalFiltration<T,co> &other) const {
    return other < *this;
  }

  inline bool operator<=(const KCriticalFiltration<T,co> &other) const {
    //check if this curves is below other's curve 
    // ie for each guy in this, check if there is a guy in other that dominates him
    for (std::size_t i = 0u; i < multi_filtration_.size(); ++i) {
      for (std::size_t j = 0u; j < other.multi_filtration_.size(); ++j) {
        // i <= j 
        if (dominates(other.multi_filtration_[j], multi_filtration_[i]))
          continue;
      }
      return false;
    }
    return true;
  }
  inline bool operator>=(const KCriticalFiltration<T,co> &other) const {
    return other <= *this;
  }

public:
  // for compiler
  constexpr static const T T_inf = Finitely_critical_multi_filtration<T>::T_inf;
  constexpr static const bool is_multi_critical = true;

private:
  std::vector<Finitely_critical_multi_filtration<T>> multi_filtration_;
};

} // namespace Gudhi::multiparameter::multi_filtrations

namespace std {

template <typename T>
class numeric_limits<Gudhi::multiparameter::multi_filtrations::
                         Finitely_critical_multi_filtration<T>> {
public:
  static constexpr bool has_infinity = std::numeric_limits<T>::has_infinity;

  static Gudhi::multiparameter::multi_filtrations::
      Finitely_critical_multi_filtration<T>
      infinity() throw() {
    return Gudhi::multiparameter::multi_filtrations::
        Finitely_critical_multi_filtration<T>(
            1, std::numeric_limits<T>::infinity());
  };
  static Gudhi::multiparameter::multi_filtrations::
      Finitely_critical_multi_filtration<T>
      minus_infinity() throw() {
    return Gudhi::multiparameter::multi_filtrations::
        Finitely_critical_multi_filtration<T>(
            1, -std::numeric_limits<T>::infinity());
  };
  static Gudhi::multiparameter::multi_filtrations::
      Finitely_critical_multi_filtration<T>
      max() throw() {
    return Gudhi::multiparameter::multi_filtrations::
        Finitely_critical_multi_filtration<T>(1, std::numeric_limits<T>::max());
  };
  static Gudhi::multiparameter::multi_filtrations::
      Finitely_critical_multi_filtration<T>
      quiet_NaN() throw() {
    return Gudhi::multiparameter::multi_filtrations::
        Finitely_critical_multi_filtration<T>(1,
                                              numeric_limits<T>::quiet_NaN());
  };
};

} // namespace std

#endif // FINITELY_CRITICAL_FILTRATIONS_H_
